"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const vscode = require("vscode");
const xslLexer_1 = require("./xslLexer");
const globalsProvider_1 = require("./globalsProvider");
const xsltTokenDefintions_1 = require("./xsltTokenDefintions");
const xsltTokenCompletions_1 = require("./xsltTokenCompletions");
const xsltSymbolProvider_1 = require("./xsltSymbolProvider");
const xpLexer_1 = require("./xpLexer");
const extension_1 = require("./extension");
const documentChangeHandler_1 = require("./documentChangeHandler");
const url = require("url");
class XsltDefinitionProvider {
    constructor(xsltConfiguration) {
        this.xpLexer = null;
        this.gp = new globalsProvider_1.GlobalsProvider();
        this.languageConfig = xsltConfiguration;
        this.xslLexer = new xslLexer_1.XslLexer(xsltConfiguration);
        this.xslLexer.provideCharLevelState = true;
        this.docType = xsltConfiguration.docType;
        this.schemaData = xsltConfiguration.schemaData;
    }
    getXPLexer() {
        if (this.xpLexer === null) {
            this.xpLexer = new xpLexer_1.XPathLexer();
        }
        return this.xpLexer;
    }
    provideDefinition(document, position, token) {
        return __awaiter(this, void 0, void 0, function* () {
            const lexPosition = { line: 0, startCharacter: 0, documentOffset: 0 };
            let extractedImportData = yield this.getImportedGlobals(document, lexPosition);
            const { allTokens, globalInstructionData, allImportedGlobals, accumulatedHrefs } = extractedImportData;
            return new Promise((resolve, reject) => {
                let location = undefined;
                let isXSLT = this.docType === xslLexer_1.DocumentTypes.XSLT;
                const defnData = xsltTokenDefintions_1.XsltTokenDefinitions.findDefinition(isXSLT, document, allTokens, globalInstructionData, allImportedGlobals, position);
                location = defnData.definitionLocation;
                if (location) {
                    location.extractedImportData = extractedImportData;
                }
                resolve(location);
            });
        });
    }
    seekDefinition(document, position, token) {
        return __awaiter(this, void 0, void 0, function* () {
            // extends provideDefinition so, if position within an instruction, it returns the instruction
            const lexPosition = { line: 0, startCharacter: 0, documentOffset: 0 };
            let extractedImportData = yield this.getImportedGlobals(document, lexPosition);
            const { allTokens, globalInstructionData, allImportedGlobals, accumulatedHrefs } = extractedImportData;
            let matchingGlobal = globalInstructionData.find(global => {
                return global.token.line === position.line &&
                    position.character >= global.token.startCharacter &&
                    position.character <= global.token.startCharacter + global.token.length;
            });
            return new Promise((resolve, reject) => {
                let location = undefined;
                let defnData = undefined;
                let isXSLT = this.docType === xslLexer_1.DocumentTypes.XSLT;
                if (matchingGlobal) {
                    location = xsltTokenDefintions_1.XsltTokenDefinitions.createLocationFromInstrcution(matchingGlobal, document);
                    defnData = { definitionLocation: location };
                }
                else {
                    defnData = xsltTokenDefintions_1.XsltTokenDefinitions.findDefinition(isXSLT, document, allTokens, globalInstructionData, allImportedGlobals, position);
                    location = defnData.definitionLocation;
                }
                if (location) {
                    location.extractedImportData = extractedImportData;
                }
                resolve(defnData);
            });
        });
    }
    static functionInstructionFromDocPosition(document, position) {
        let fnName;
        let fnInstruction;
        const line = document.lineAt(position.line).text;
        let startIndex = 0;
        for (startIndex = position.character; startIndex > 0; startIndex--) {
            const c = line.charAt(startIndex);
            if (!XsltDefinitionProvider.nameCharRgx.test(c)) {
                break;
            }
        }
        let endIndex = 0;
        let isFunction = false;
        for (endIndex = position.character + 1; endIndex < line.length; endIndex++) {
            const c = line.charAt(endIndex);
            if (!XsltDefinitionProvider.nameCharRgx.test(c)) {
                if (c === '(') {
                    isFunction = true;
                }
                break;
            }
        }
        if (isFunction) {
            fnName = line.substring(startIndex + 1, endIndex);
            const token = { line: position.line, startCharacter: startIndex + 1, length: endIndex - (startIndex + 1), tokenType: 0, value: fnName };
            fnInstruction = { token: token, idNumber: -1, name: fnName, type: xslLexer_1.GlobalInstructionType.Function, href: url.fileURLToPath(document.uri.toString()) };
        }
        return fnInstruction;
    }
    getImportedGlobals(document, lexPosition) {
        return __awaiter(this, void 0, void 0, function* () {
            let allTokens = [];
            let globalInstructionData = [];
            if (this.docType === xslLexer_1.DocumentTypes.XPath) {
                allTokens = this.getXPLexer().analyse(document.getText(), xpLexer_1.ExitCondition.None, lexPosition);
                globalInstructionData = extension_1.XPathSemanticTokensProvider.getGlobalInstructionData();
            }
            else {
                allTokens = this.xslLexer.analyse(document.getText());
                globalInstructionData = this.xslLexer.globalInstructionData;
            }
            const xsltPackages = vscode.workspace.getConfiguration('XSLT.resources').get('xsltPackages');
            // Import/include XSLT - ensuring no duplicates
            const localImportedHrefs = xsltSymbolProvider_1.XsltSymbolProvider.importSymbolHrefs;
            let { importedGlobals1, accumulatedHrefs } = yield xsltSymbolProvider_1.XsltSymbolProvider.processTopLevelImports(false, this.xslLexer, localImportedHrefs, document, globalInstructionData, xsltPackages);
            let globalsSummary0 = { globals: importedGlobals1, hrefs: accumulatedHrefs };
            const maxImportLevel = 20;
            let processNestedGlobals = () => __awaiter(this, void 0, void 0, function* () {
                let level = 0;
                while (globalsSummary0.hrefs.length > 0 && level < maxImportLevel) {
                    globalsSummary0 = yield xsltSymbolProvider_1.XsltSymbolProvider.processImportedGlobals(xsltPackages, globalsSummary0.globals, accumulatedHrefs, level === 0);
                    level++;
                }
            });
            yield processNestedGlobals();
            let allImportedGlobals = [];
            globalsSummary0.globals.forEach((globals) => {
                if (globals.error) {
                    // ignore 
                }
                else {
                    globals.data.forEach((global) => {
                        global['href'] = globals.href;
                        allImportedGlobals.push(global);
                    });
                }
            });
            return { allTokens, globalInstructionData, allImportedGlobals, accumulatedHrefs };
        });
    }
    provideCompletionItems(document, position, token, context) {
        var _a, _b, _c;
        return __awaiter(this, void 0, void 0, function* () {
            const keepNameTests = true;
            const lexPosition = { line: 0, startCharacter: 0, documentOffset: 0 };
            let symbolsForXPath = [];
            let allTokens = [];
            let globalInstructionData = [];
            let uri;
            if (this.docType === xslLexer_1.DocumentTypes.XPath) {
                allTokens = this.getXPLexer().analyse(document.getText(), xpLexer_1.ExitCondition.None, lexPosition);
                globalInstructionData = extension_1.XPathSemanticTokensProvider.getGlobalInstructionData();
                uri = (_a = documentChangeHandler_1.DocumentChangeHandler.lastActiveXMLEditor) === null || _a === void 0 ? void 0 : _a.document.uri;
            }
            else {
                if (this.docType === xslLexer_1.DocumentTypes.XSLT) {
                    uri = (_b = documentChangeHandler_1.DocumentChangeHandler.lastActiveXMLNonXSLEditor) === null || _b === void 0 ? void 0 : _b.document.uri;
                }
                else {
                    uri = (_c = documentChangeHandler_1.DocumentChangeHandler.lastActiveXMLEditor) === null || _c === void 0 ? void 0 : _c.document.uri;
                }
                allTokens = this.xslLexer.analyse(document.getText(), keepNameTests);
                globalInstructionData = this.xslLexer.globalInstructionData;
            }
            if (uri) {
                const lastSymbols = xsltSymbolProvider_1.XsltSymbolProvider.documentSymbols.get(uri);
                if (lastSymbols) {
                    symbolsForXPath = lastSymbols;
                }
            }
            const xsltPackages = vscode.workspace.getConfiguration('XSLT.resources').get('xsltPackages');
            // Import/include XSLT - ensuring no duplicates
            const localImportedHrefs = xsltSymbolProvider_1.XsltSymbolProvider.importSymbolHrefs;
            let { importedGlobals1, accumulatedHrefs } = yield xsltSymbolProvider_1.XsltSymbolProvider.processTopLevelImports(false, this.xslLexer, localImportedHrefs, document, globalInstructionData, xsltPackages);
            let globalsSummary0 = { globals: importedGlobals1, hrefs: accumulatedHrefs };
            const maxImportLevel = 20;
            let processNestedGlobals = () => __awaiter(this, void 0, void 0, function* () {
                let level = 0;
                while (globalsSummary0.hrefs.length > 0 && level < maxImportLevel) {
                    globalsSummary0 = yield xsltSymbolProvider_1.XsltSymbolProvider.processImportedGlobals(xsltPackages, globalsSummary0.globals, accumulatedHrefs, level === 0);
                    level++;
                }
            });
            yield processNestedGlobals();
            return new Promise((resolve, reject) => {
                let allImportedGlobals = [];
                globalsSummary0.globals.forEach((globals) => {
                    if (globals.error) {
                        // ignore 
                    }
                    else {
                        globals.data.forEach((global) => {
                            global['href'] = globals.href;
                            allImportedGlobals.push(global);
                        });
                    }
                });
                let attNames = this.xslLexer.attributeNameTests ? this.xslLexer.attributeNameTests : [];
                let nodeNames = this.xslLexer.elementNameTests ? this.xslLexer.elementNameTests : [];
                let xslVariable = ['xsl:variable', 'xsl:param'];
                let completions;
                completions = xsltTokenCompletions_1.XsltTokenCompletions.getCompletions(this.languageConfig, symbolsForXPath, xslVariable, attNames, nodeNames, document, allTokens, globalInstructionData, allImportedGlobals, position);
                resolve(completions);
            });
        });
    }
}
exports.XsltDefinitionProvider = XsltDefinitionProvider;
XsltDefinitionProvider.nameCharRgx = new RegExp(/[A-Z]|[a-z]|_|-|:/);
//# sourceMappingURL=xsltDefinitionProvider.js.map