"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const vscode = require("vscode");
const xslLexer_1 = require("./xslLexer");
const xsltTokenDiagnostics_1 = require("./xsltTokenDiagnostics");
const globalsProvider_1 = require("./globalsProvider");
const path = require("path");
const documentChangeHandler_1 = require("./documentChangeHandler");
var SelectionType;
(function (SelectionType) {
    SelectionType[SelectionType["Current"] = 0] = "Current";
    SelectionType[SelectionType["Next"] = 1] = "Next";
    SelectionType[SelectionType["Previous"] = 2] = "Previous";
    SelectionType[SelectionType["Parent"] = 3] = "Parent";
})(SelectionType = exports.SelectionType || (exports.SelectionType = {}));
class XsltSymbolProvider {
    constructor(xsltConfiguration, collection) {
        this.importHrefs = new Map();
        this.xslLexer = new xslLexer_1.XslLexer(xsltConfiguration);
        this.xslLexer.provideCharLevelState = true;
        this.collection = collection;
        this.languageConfig = xsltConfiguration;
        this.docType = xsltConfiguration.docType;
    }
    provideDocumentSymbols(document, token) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = this.getDocumentSymbols(document, true);
            return result;
        });
    }
    static findMatchingParent(map, path) {
        let matchingKey;
        // use last entry
        for (const entry of map.entries()) {
            if (entry[1].find((x) => x === path)) {
                matchingKey = entry[0];
            }
        }
        return matchingKey;
    }
    getDocumentSymbols(document, onlyActiveDocument) {
        return __awaiter(this, void 0, void 0, function* () {
            XsltSymbolProvider.documentSymbols = [];
            if (onlyActiveDocument && vscode.window.activeTextEditor) {
                if (document.fileName !== vscode.window.activeTextEditor.document.fileName) {
                    return [];
                }
            }
            const allTokens = this.xslLexer.analyse(document.getText());
            const globalInstructionData = this.xslLexer.globalInstructionData;
            const xsltPackages = vscode.workspace.getConfiguration('XSLT.resources').get('xsltPackages');
            // Import/include XSLT - ensuring no duplicates
            const localImportedHrefs = XsltSymbolProvider.importSymbolHrefs;
            let { importedGlobals1, accumulatedHrefs } = yield XsltSymbolProvider.processTopLevelImports(true, this.xslLexer, localImportedHrefs, document, globalInstructionData, xsltPackages);
            let topLevelHrefs = XsltSymbolProvider.accumulateImportHrefs(xsltPackages, importedGlobals1, []);
            let globalsSummary0 = { globals: importedGlobals1, hrefs: accumulatedHrefs };
            const maxImportLevel = 20;
            let processNestedGlobals = () => __awaiter(this, void 0, void 0, function* () {
                let level = 0;
                while (globalsSummary0.hrefs.length > 0 && level < maxImportLevel) {
                    globalsSummary0 = yield XsltSymbolProvider.processImportedGlobals(xsltPackages, globalsSummary0.globals, accumulatedHrefs, level === 0);
                    level++;
                }
            });
            yield processNestedGlobals();
            return new Promise((resolve, reject) => {
                let symbols = [];
                let allImportedGlobals = [];
                let importErrors = [];
                const rootPath = vscode.workspace.rootPath;
                globalsSummary0.globals.forEach((globals) => {
                    if (globals.error) {
                        if (topLevelHrefs.indexOf(globals.href) > -1) {
                            let errorData = globalInstructionData.find((dataObject) => {
                                let result = false;
                                if (dataObject.type === xslLexer_1.GlobalInstructionType.Import || dataObject.type === xslLexer_1.GlobalInstructionType.Include) {
                                    let resolvedName = XsltSymbolProvider.resolvePath(dataObject.name, document.fileName);
                                    result = resolvedName === globals.href;
                                }
                                else if (dataObject.type === xslLexer_1.GlobalInstructionType.UsePackage) {
                                    // TODO:
                                    const basePath = path.dirname(document.fileName);
                                    let packageLookup = xsltPackages.find((pkg) => {
                                        return pkg.name === dataObject.name;
                                    });
                                    if (packageLookup && rootPath) {
                                        let resolvedName = XsltSymbolProvider.resolvePathInSettings(packageLookup.path, rootPath);
                                        result = resolvedName === globals.href;
                                    }
                                    else {
                                        result = false;
                                    }
                                }
                                return result;
                            });
                            if (errorData) {
                                importErrors.push(errorData);
                            }
                        }
                    }
                    else {
                        globals.data.forEach((global) => {
                            global['href'] = globals.href;
                            allImportedGlobals.push(global);
                        });
                    }
                });
                let diagnostics = xsltTokenDiagnostics_1.XsltTokenDiagnostics.calculateDiagnostics(this.languageConfig, this.docType, document, allTokens, globalInstructionData, allImportedGlobals, symbols);
                // TODO: don't exclude DocumentTypes.SCH once diagnostics implemented fully
                if (this.collection && this.docType !== xslLexer_1.DocumentTypes.SCH) {
                    let importDiagnostics = [];
                    importErrors.forEach((importError) => {
                        importDiagnostics.push(xsltTokenDiagnostics_1.XsltTokenDiagnostics.createImportDiagnostic(importError));
                    });
                    let allDiagnostics = importDiagnostics.concat(diagnostics);
                    if (allDiagnostics.length > 0) {
                        this.collection.set(document.uri, allDiagnostics);
                    }
                    else {
                        this.collection.clear();
                    }
                    ;
                }
                XsltSymbolProvider.documentSymbols = symbols;
                resolve(symbols);
            });
        });
    }
    static processTopLevelImports(update, xslLexer, localImportedHrefs, document, globalInstructionData, xsltPackages) {
        return __awaiter(this, void 0, void 0, function* () {
            const matchingParent = this.findMatchingParent(localImportedHrefs, document.fileName);
            let importedGlobals1 = [];
            let accumulatedHrefs;
            if (matchingParent) {
                // TODO: get globalInstructionData for this point!
                const token = {
                    line: 1,
                    startCharacter: 0,
                    length: matchingParent === null || matchingParent === void 0 ? void 0 : matchingParent.length,
                    value: matchingParent,
                    tokenType: 0
                };
                const importInstruction = {
                    type: xslLexer_1.GlobalInstructionType.Import,
                    name: matchingParent,
                    token: token,
                    idNumber: 0
                };
                globalInstructionData.push(importInstruction);
            }
            const importedG = { data: globalInstructionData, href: document.fileName, error: false };
            accumulatedHrefs = [importedG.href];
            importedGlobals1 = [importedG];
            const inheritHrefs = this.getImportHrefs(xsltPackages, importedGlobals1);
            if (update && !matchingParent) {
                localImportedHrefs.set(document.fileName, inheritHrefs);
            }
            return { importedGlobals1, accumulatedHrefs };
        });
    }
    static selectTextWithSymbol(symbol) {
        if (documentChangeHandler_1.DocumentChangeHandler.lastActiveXMLEditor && symbol) {
            const range = symbol.range;
            documentChangeHandler_1.DocumentChangeHandler.lastActiveXMLEditor.selection = new vscode.Selection(range.start, range.end);
        }
    }
    static selectXMLElement(selectionType) {
        var _a;
        const selection = (_a = vscode.window.activeTextEditor) === null || _a === void 0 ? void 0 : _a.selection;
        if (vscode.window.activeTextEditor && selection) {
            const rootSymbol = XsltSymbolProvider.documentSymbols[0];
            const path = [];
            const result = this.getChildSymbolForSelection(selection, rootSymbol, path, selectionType, null, null, null);
            if (result !== null) {
                XsltSymbolProvider.selectTextWithSymbol(result);
            }
        }
    }
    static getXPathFromSelection() {
        const editor = vscode.window.activeTextEditor;
        if (editor) {
            const selection = editor.selection;
            const rootSymbol = XsltSymbolProvider.documentSymbols[0];
            const newPath = ['/' + rootSymbol.name.split(' ')[0]];
            const result = this.getChildSymbolForSelection(selection, rootSymbol, newPath, SelectionType.Current, null, null, null);
            const fullPath = newPath.join('');
            return fullPath;
        }
    }
    static getChildSymbolForSelection(selection, symbol, path, selectionType, parentSymbol, precedingSymbol, nextSymbol) {
        const result = symbol.children.find((sym) => {
            const selectionPos = new vscode.Position(selection.start.line, selection.start.character);
            return sym.range.contains(selectionPos);
        });
        if (result) {
            const resultName = result.name.split(' ')[0];
            let nextSibling = null;
            let precedingSibling = null;
            let precedingSymbolNames = 1;
            let breakNext = false;
            for (const sibling of symbol.children) {
                if (breakNext) {
                    nextSibling = sibling;
                    break;
                }
                if (sibling === result) {
                    if (selectionType === SelectionType.Next) {
                        breakNext = true;
                    }
                    else {
                        break;
                    }
                }
                else {
                    precedingSibling = sibling;
                    const siblingName = sibling.name.split(' ')[0];
                    if (siblingName === resultName) {
                        precedingSymbolNames++;
                    }
                }
            }
            if (symbol.kind === vscode.SymbolKind.Array && symbol.name === 'attributes') {
                path.push('/@' + resultName);
            }
            else if (!(result.kind === vscode.SymbolKind.Array && result.name === 'attributes')) {
                path.push('/' + resultName + `[${precedingSymbolNames}]`);
            }
            return this.getChildSymbolForSelection(selection, result, path, selectionType, symbol, precedingSibling, nextSibling);
        }
        else if (selectionType === SelectionType.Parent) {
            return parentSymbol;
        }
        else if (selectionType === SelectionType.Previous) {
            return precedingSymbol;
        }
        else if (selectionType === SelectionType.Next) {
            return nextSymbol;
        }
        else {
            return symbol;
        }
    }
    static getSymbolFromXPathLocator(rawText, symbols) {
        const text = rawText.startsWith('/') ? rawText.substring(1) : '';
        const pathParts = text.split('/');
        if (symbols.length === 0) {
            return;
        }
        let currentSymbol = symbols[0];
        pathParts.forEach((item, i) => {
            const isLastItem = i === pathParts.length - 1;
            if (currentSymbol === undefined) {
                // can't yet handle attribute-test: /element/@attribute
                return;
            }
            else if (isLastItem && item.startsWith('@') && currentSymbol.children.length > 0) {
                const attrName = item.substring(1);
                const attributesSymbol = undefined;
                const firstChild = currentSymbol.children[0];
                if (firstChild.kind === vscode.SymbolKind.Array && firstChild.name === 'attributes') {
                    currentSymbol = firstChild.children.find(symbol => symbol.name === attrName);
                }
                else {
                    currentSymbol = undefined;
                }
                return;
            }
            const pos = item.indexOf('[');
            let pathName = '';
            // only want one result;
            let pathIndex = 1;
            if (pos === -1) {
                pathName = item;
            }
            else {
                const predicateString = item.substring(pos + 1, item.length - 1);
                const predicateNum = Number(predicateString);
                pathIndex = -1;
                if (predicateNum !== NaN) {
                    pathIndex = predicateNum;
                }
                pathName = item.substr(0, pos);
            }
            if (i === 0) {
                const spacePos = currentSymbol.name.indexOf(' ');
                const symbolName = spacePos === -1 ? currentSymbol.name : currentSymbol.name.substring(0, spacePos);
                currentSymbol = symbolName === pathName && pathIndex === 1 ? currentSymbol : undefined;
            }
            else {
                let nameCount = 0;
                currentSymbol = currentSymbol.children.find((symbol) => {
                    const spacePos = symbol.name.indexOf(' ');
                    const symbolName = spacePos === -1 ? symbol.name : symbol.name.substring(0, spacePos);
                    if (symbolName === pathName) {
                        nameCount++;
                        return pathIndex === nameCount;
                    }
                    else {
                        return false;
                    }
                });
            }
        });
        return currentSymbol;
    }
    static processImportedGlobals(xsltPackages, importedGlobals1, level1Hrefs, topLevel) {
        return __awaiter(this, void 0, void 0, function* () {
            let level2Globals = [];
            let level2Hrefs = XsltSymbolProvider.accumulateImportHrefs(xsltPackages, importedGlobals1, level1Hrefs);
            let newGlobals = [];
            level2Hrefs.forEach((href) => {
                level2Globals.push(XsltSymbolProvider.fetchImportedGlobals([href]));
            });
            let importedGlobals2Array = yield Promise.all(level2Globals);
            importedGlobals2Array.forEach((importedGlobals2) => {
                if (topLevel) {
                    newGlobals = newGlobals.concat(importedGlobals2);
                }
                else {
                    importedGlobals1 = importedGlobals1.concat(importedGlobals2);
                }
            });
            if (topLevel) {
                return { globals: newGlobals, hrefs: level2Hrefs };
            }
            else {
                return { globals: importedGlobals1, hrefs: level2Hrefs };
            }
        });
    }
    static accumulateImportHrefs(xsltPackages, importedGlobals, existingHrefs) {
        let result = [];
        const rootPath = vscode.workspace.rootPath;
        importedGlobals.forEach((importedG) => {
            importedG.data.forEach((data) => {
                if (data.type === xslLexer_1.GlobalInstructionType.Import || data.type === xslLexer_1.GlobalInstructionType.Include) {
                    let resolvedName = XsltSymbolProvider.resolvePath(data.name, importedG.href);
                    if (existingHrefs.indexOf(resolvedName) < 0) {
                        existingHrefs.push(resolvedName);
                        result.push(resolvedName);
                    }
                }
                else if (rootPath && data.type === xslLexer_1.GlobalInstructionType.UsePackage) {
                    let packageLookup = xsltPackages.find((pkg) => {
                        return pkg.name === data.name;
                    });
                    if (packageLookup) {
                        let resolvedName = XsltSymbolProvider.resolvePathInSettings(packageLookup.path, rootPath);
                        if (existingHrefs.indexOf(resolvedName) < 0) {
                            existingHrefs.push(resolvedName);
                            result.push(resolvedName);
                        }
                    }
                }
            });
        });
        return result;
    }
    static getImportHrefs(xsltPackages, importedGlobals) {
        let result = [];
        const rootPath = vscode.workspace.rootPath;
        importedGlobals.forEach((importedG) => {
            importedG.data.forEach((data) => {
                if (data.type === xslLexer_1.GlobalInstructionType.Import || data.type === xslLexer_1.GlobalInstructionType.Include) {
                    let resolvedName = XsltSymbolProvider.resolvePath(data.name, importedG.href);
                    result.push(resolvedName);
                }
                else if (rootPath && data.type === xslLexer_1.GlobalInstructionType.UsePackage) {
                    let packageLookup = xsltPackages.find((pkg) => {
                        return pkg.name === data.name;
                    });
                    if (packageLookup) {
                        let resolvedName = XsltSymbolProvider.resolvePathInSettings(packageLookup.path, rootPath);
                        result.push(resolvedName);
                    }
                }
            });
        });
        return result;
    }
    static resolvePath(href, documentPath) {
        if (path.isAbsolute(href)) {
            return href;
        }
        else if (href.startsWith('file:///')) {
            return href.substring(7);
        }
        else {
            href = href.startsWith('file:') ? href.substring(5) : href;
            let basePath = path.dirname(documentPath);
            let joinedPath = path.join(basePath, href);
            return path.normalize(joinedPath);
        }
    }
    static resolvePathInSettings(href, workspace) {
        if (path.isAbsolute(href)) {
            return href;
        }
        else {
            let joinedPath = path.join(workspace, href);
            return path.normalize(joinedPath);
        }
    }
    static fetchImportedGlobals(inputHrefs) {
        return __awaiter(this, void 0, void 0, function* () {
            let result = [];
            //let inputHrefs: string[] = this.accumulateImportHrefs(globalInstructionData, existingHrefs, docHref);
            let lastIndex = inputHrefs.length - 1;
            if (lastIndex < 0) {
                return result;
            }
            else {
                return new Promise((resolve, reject) => {
                    inputHrefs.forEach((href, index) => {
                        globalsProvider_1.GlobalsProvider.provideGlobals(href).then((globals) => {
                            result.push({ href: href, data: globals.data, error: globals.error });
                            if (index === lastIndex) {
                                resolve(result);
                            }
                        });
                    });
                });
            }
        });
    }
}
exports.XsltSymbolProvider = XsltSymbolProvider;
XsltSymbolProvider.documentSymbols = [];
XsltSymbolProvider.importSymbolHrefs = new Map();
//# sourceMappingURL=xsltDiagnosticsProvider.js.map