"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const vscode = require("vscode");
const xslLexer_1 = require("./xslLexer");
const xsltTokenDiagnostics_1 = require("./xsltTokenDiagnostics");
const globalsProvider_1 = require("./globalsProvider");
const path = require("path");
const documentChangeHandler_1 = require("./documentChangeHandler");
const xpLexer_1 = require("./xpLexer");
const xsltTokenCompletions_1 = require("./xsltTokenCompletions");
var SelectionType;
(function (SelectionType) {
    SelectionType[SelectionType["Current"] = 0] = "Current";
    SelectionType[SelectionType["Next"] = 1] = "Next";
    SelectionType[SelectionType["Previous"] = 2] = "Previous";
    SelectionType[SelectionType["Parent"] = 3] = "Parent";
    SelectionType[SelectionType["FirstChild"] = 4] = "FirstChild";
})(SelectionType = exports.SelectionType || (exports.SelectionType = {}));
var AxisType;
(function (AxisType) {
    AxisType[AxisType["Attribute"] = 0] = "Attribute";
    AxisType[AxisType["Self"] = 1] = "Self";
    AxisType[AxisType["Child"] = 2] = "Child";
    AxisType[AxisType["Descendant"] = 3] = "Descendant";
    AxisType[AxisType["DescendantOrSelf"] = 4] = "DescendantOrSelf";
    AxisType[AxisType["Parent"] = 5] = "Parent";
    AxisType[AxisType["ParentKeep"] = 6] = "ParentKeep";
    AxisType[AxisType["Ancestor"] = 7] = "Ancestor";
    AxisType[AxisType["AncestorOrSelf"] = 8] = "AncestorOrSelf";
    AxisType[AxisType["FollowingSibling"] = 9] = "FollowingSibling";
    AxisType[AxisType["PrecedingSibling"] = 10] = "PrecedingSibling";
    AxisType[AxisType["Union"] = 11] = "Union";
})(AxisType || (AxisType = {}));
var UseSource;
(function (UseSource) {
    UseSource["none"] = "none";
    UseSource["latest"] = "latest";
})(UseSource || (UseSource = {}));
class XsltSymbolProvider {
    constructor(xsltConfiguration, collection) {
        this.importHrefs = new Map();
        this.xslLexer = new xslLexer_1.XslLexer(xsltConfiguration);
        this.xslLexer.provideCharLevelState = true;
        this.collection = collection;
        this.languageConfig = xsltConfiguration;
        this.docType = xsltConfiguration.docType;
    }
    static getSymbolsForActiveDocument() {
        if (vscode.window.activeTextEditor) {
            const result = XsltSymbolProvider.documentSymbols.get(vscode.window.activeTextEditor.document.uri);
            return result ? result : [];
        }
        else {
            return [];
        }
    }
    provideDocumentSymbols(document, token) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = this.getDocumentSymbols(document, false);
            return result;
        });
    }
    static findMatchingParent(map, path) {
        let matchingKey;
        // use last entry
        for (const entry of map.entries()) {
            if (entry[1].find((x) => x === path)) {
                matchingKey = entry[0];
            }
        }
        return matchingKey;
    }
    getDocumentSymbols(document, onlyActiveDocument) {
        return __awaiter(this, void 0, void 0, function* () {
            if (onlyActiveDocument && vscode.window.activeTextEditor) {
                if (document.fileName !== vscode.window.activeTextEditor.document.fileName) {
                    return [];
                }
            }
            const allTokens = this.xslLexer.analyse(document.getText());
            const globalInstructionData = this.xslLexer.globalInstructionData;
            const xsltPackages = vscode.workspace.getConfiguration('XSLT.resources').get('xsltPackages');
            // Import/include XSLT - ensuring no duplicates
            const localImportedHrefs = XsltSymbolProvider.importSymbolHrefs;
            let { importedGlobals1, accumulatedHrefs } = yield XsltSymbolProvider.processTopLevelImports(true, this.xslLexer, localImportedHrefs, document, globalInstructionData, xsltPackages);
            let topLevelHrefs = XsltSymbolProvider.accumulateImportHrefs(xsltPackages, importedGlobals1, []);
            let globalsSummary0 = { globals: importedGlobals1, hrefs: accumulatedHrefs };
            const maxImportLevel = 20;
            let processNestedGlobals = () => __awaiter(this, void 0, void 0, function* () {
                let level = 0;
                while (globalsSummary0.hrefs.length > 0 && level < maxImportLevel) {
                    globalsSummary0 = yield XsltSymbolProvider.processImportedGlobals(xsltPackages, globalsSummary0.globals, accumulatedHrefs, level === 0);
                    level++;
                }
            });
            yield processNestedGlobals();
            return new Promise((resolve, reject) => {
                let symbols = [];
                let allImportedGlobals = [];
                let importErrors = [];
                const rootPath = vscode.workspace.rootPath;
                globalsSummary0.globals.forEach((globals) => {
                    if (globals.error) {
                        if (topLevelHrefs.indexOf(globals.href) > -1) {
                            let errorData = globalInstructionData.find((dataObject) => {
                                let result = false;
                                if (dataObject.type === xslLexer_1.GlobalInstructionType.Import || dataObject.type === xslLexer_1.GlobalInstructionType.Include) {
                                    let resolvedName = XsltSymbolProvider.resolvePath(dataObject.name, document.fileName);
                                    result = resolvedName === globals.href;
                                }
                                else if (dataObject.type === xslLexer_1.GlobalInstructionType.UsePackage) {
                                    // TODO:
                                    const basePath = path.dirname(document.fileName);
                                    let packageLookup = xsltPackages.find((pkg) => {
                                        return pkg.name === dataObject.name;
                                    });
                                    if (packageLookup && rootPath) {
                                        let resolvedName = XsltSymbolProvider.resolvePathInSettings(packageLookup.path, rootPath);
                                        result = resolvedName === globals.href;
                                    }
                                    else {
                                        result = false;
                                    }
                                }
                                return result;
                            });
                            if (errorData) {
                                importErrors.push(errorData);
                            }
                        }
                    }
                    else {
                        globals.data.forEach((global) => {
                            global['href'] = globals.href;
                            allImportedGlobals.push(global);
                        });
                    }
                });
                let diagnostics = xsltTokenDiagnostics_1.XsltTokenDiagnostics.calculateDiagnostics(this.languageConfig, this.docType, document, allTokens, globalInstructionData, allImportedGlobals, symbols);
                if (vscode.window.activeTextEditor && document.fileName !== vscode.window.activeTextEditor.document.fileName) {
                }
                else {
                    // TODO: don't exclude DocumentTypes.SCH once diagnostics implemented fully
                    if (this.collection && this.docType !== xslLexer_1.DocumentTypes.SCH) {
                        let importDiagnostics = [];
                        importErrors.forEach((importError) => {
                            importDiagnostics.push(xsltTokenDiagnostics_1.XsltTokenDiagnostics.createImportDiagnostic(importError));
                        });
                        let allDiagnostics = importDiagnostics.concat(diagnostics);
                        if (allDiagnostics.length > 0) {
                            this.collection.set(document.uri, allDiagnostics);
                        }
                        else {
                            this.collection.clear();
                        }
                        ;
                    }
                }
                XsltSymbolProvider.documentSymbols.set(document.uri, symbols);
                resolve(symbols);
            });
        });
    }
    static processTopLevelImports(update, xslLexer, localImportedHrefs, document, globalInstructionData, xsltPackages) {
        var _a;
        return __awaiter(this, void 0, void 0, function* () {
            const matchingParent = this.findMatchingParent(localImportedHrefs, document.fileName);
            let importedGlobals1 = [];
            let accumulatedHrefs;
            if (matchingParent) {
                // TODO: get globalInstructionData for this point!
                const token = {
                    line: 1,
                    startCharacter: 0,
                    length: matchingParent === null || matchingParent === void 0 ? void 0 : matchingParent.length,
                    value: matchingParent,
                    tokenType: 0
                };
                const importInstruction = {
                    type: xslLexer_1.GlobalInstructionType.Import,
                    name: matchingParent,
                    token: token,
                    idNumber: 0
                };
                globalInstructionData.push(importInstruction);
            }
            const importedG = { data: globalInstructionData, href: document.fileName, error: false };
            accumulatedHrefs = [importedG.href];
            importedGlobals1 = [importedG];
            const inheritHrefs = this.getImportHrefs(xsltPackages, importedGlobals1);
            if (update) {
                if (matchingParent) {
                    const newInheritHres = (_a = localImportedHrefs.get(matchingParent)) === null || _a === void 0 ? void 0 : _a.concat(inheritHrefs);
                    if (newInheritHres) {
                        localImportedHrefs.set(matchingParent, newInheritHres);
                    }
                }
                else {
                    localImportedHrefs.set(document.fileName, inheritHrefs);
                }
            }
            return { importedGlobals1, accumulatedHrefs };
        });
    }
    static selectTextWithSymbol(symbol) {
        if (documentChangeHandler_1.DocumentChangeHandler.lastActiveXMLEditor && symbol) {
            const range = symbol.range;
            if (vscode.window.activeTextEditor) {
                vscode.window.activeTextEditor.revealRange(range);
            }
            documentChangeHandler_1.DocumentChangeHandler.lastActiveXMLEditor.selection = new vscode.Selection(range.start, range.end);
        }
    }
    static selectXMLElement(selectionType) {
        var _a;
        const selection = (_a = vscode.window.activeTextEditor) === null || _a === void 0 ? void 0 : _a.selection;
        if (vscode.window.activeTextEditor && selection) {
            const rootSymbol = XsltSymbolProvider.getSymbolsForActiveDocument()[0];
            const path = [];
            const result = this.getChildSymbolForSelection(selection, rootSymbol, path, selectionType, null, null, null);
            if (result !== null) {
                XsltSymbolProvider.selectTextWithSymbol(result);
            }
        }
    }
    static getXPathFromSelection() {
        const editor = vscode.window.activeTextEditor;
        if (editor) {
            const selection = editor.selection;
            const rootSymbol = XsltSymbolProvider.getSymbolsForActiveDocument()[0];
            const newPath = ['/' + rootSymbol.name.split(' ')[0]];
            const result = this.getChildSymbolForSelection(selection, rootSymbol, newPath, SelectionType.Current, null, null, null);
            const fullPath = newPath.join('');
            return fullPath;
        }
    }
    static getChildSymbolForSelection(selection, symbol, path, selectionType, parentSymbol, precedingSymbol, nextSymbol) {
        const result = symbol.children.find((sym) => {
            const selectionPos = new vscode.Position(selection.start.line, selection.start.character);
            return sym.range.contains(selectionPos);
        });
        if (result) {
            const resultName = result.name.split(' ')[0];
            let nextSibling = null;
            let precedingSibling = null;
            let firstChild = null;
            let precedingSymbolNames = 1;
            let breakNext = false;
            for (const sibling of symbol.children) {
                if (breakNext) {
                    nextSibling = sibling;
                    break;
                }
                if (sibling === result) {
                    if (selectionType === SelectionType.Next) {
                        breakNext = true;
                    }
                    else {
                        break;
                    }
                }
                else {
                    precedingSibling = sibling;
                    const siblingName = sibling.name.split(' ')[0];
                    if (siblingName === resultName) {
                        precedingSymbolNames++;
                    }
                }
            }
            if (symbol.kind === vscode.SymbolKind.Array && symbol.name === 'attributes') {
                path.push('/@' + resultName);
            }
            else if (!(result.kind === vscode.SymbolKind.Array && result.name === 'attributes')) {
                path.push('/' + resultName + `[${precedingSymbolNames}]`);
            }
            return this.getChildSymbolForSelection(selection, result, path, selectionType, symbol, precedingSibling, nextSibling);
        }
        else if (selectionType === SelectionType.Parent) {
            return parentSymbol;
        }
        else if (selectionType === SelectionType.Previous) {
            return precedingSymbol;
        }
        else if (selectionType === SelectionType.Next) {
            return nextSymbol;
        }
        else if (selectionType === SelectionType.FirstChild) {
            const firstChild = symbol.children.length > 0 ? symbol.children[0] : null;
            if (firstChild) {
                if (firstChild.kind === vscode.SymbolKind.Array && firstChild.name === 'attributes') {
                    return symbol.children.length > 1 ? symbol.children[1] : null;
                }
                else {
                    return firstChild;
                }
            }
            else {
                return null;
            }
        }
        else {
            return symbol;
        }
    }
    static getCompletionNodeNames(allTokens, globalInstructionData, xsltVariables, xpathVariables, index, xpathStack, symbols, eNames, aNames) {
        if (this.useSourceFile === UseSource.none) {
            return [eNames, aNames];
        }
        const { cleanedTokens, hasParentAxis } = XsltSymbolProvider.filterPathTokens(allTokens, globalInstructionData, xsltVariables, xpathVariables, index, xpathStack);
        // console.log('pathTokens')
        // console.log(cleanedTokens);
        const result = XsltSymbolProvider.getExpectedForXPathLocation(cleanedTokens, symbols, hasParentAxis);
        const [elementNames, attrNames] = result;
        // console.log('elementNames');
        // console.log(elementNames);
        // console.log('attrtNames');
        // console.log(attrNames);
        return result;
    }
    static filterPathTokens(tokens, globalInstructionData, xsltVariables, xpathVariables, position, xpathStack) {
        let cleanedTokens = [];
        const bracketTokens = [];
        let saveToken = false;
        let nesting = 0;
        let exitLoop = false;
        let finalSlashToken;
        let xpathStacksChecked = 0;
        let lastTokenWasSlash = false;
        let lastTokenWasSlash2 = false;
        let hasParentAxis = false;
        let foundVariableToken = false;
        let isNextBracketUnion = false;
        let collectBracketTokens = false;
        // track backwards to get all tokens in path
        for (let i = position; i > -1; i--) {
            const token = tokens[i];
            const isXSLToken = token.tokenType >= xsltTokenCompletions_1.XsltTokenCompletions.xsltStartTokenNumber;
            let xpathCharType = token.charType;
            let xpathTokenType = token.tokenType;
            if (nesting > 0) {
                if (xpathTokenType === xpLexer_1.TokenLevelState.operator && xpathCharType === xpLexer_1.CharLevelState.lPr) {
                    nesting--;
                }
            }
            else if (collectBracketTokens) {
                // test to see if: (element1|element2)
                switch (xpathTokenType) {
                    case xpLexer_1.TokenLevelState.operator:
                        switch (xpathCharType) {
                            case xpLexer_1.CharLevelState.lB:
                                // TODO: handle case of: /countries/(sibling|sibling2)
                                cleanedTokens = cleanedTokens.concat(...bracketTokens);
                                // saveToken = true;
                                // temp fix to terminate when finding (...)/country
                                bracketTokens.length = 0;
                                isNextBracketUnion = false;
                                collectBracketTokens = false;
                                break;
                            case xpLexer_1.CharLevelState.sep:
                                exitLoop = (token.value !== '|' && token.value !== ',');
                                isNextBracketUnion = !exitLoop;
                                break;
                            case xpLexer_1.CharLevelState.rPr:
                                nesting++;
                                break;
                            default:
                                exitLoop = true;
                        }
                        break;
                    case xpLexer_1.TokenLevelState.nodeNameTest:
                        if (isNextBracketUnion) {
                            token.union = true;
                        }
                        bracketTokens.push(token);
                        break;
                    default:
                        exitLoop = true;
                        break;
                }
            }
            else {
                switch (xpathTokenType) {
                    case xpLexer_1.TokenLevelState.operator:
                        switch (xpathCharType) {
                            case xpLexer_1.CharLevelState.rPr:
                                nesting++;
                                break;
                            case xpLexer_1.CharLevelState.rB:
                                collectBracketTokens = true;
                                break;
                            case xpLexer_1.CharLevelState.dSep:
                                exitLoop = (token.value !== '//' && token.value !== '::' && token.value !== '()');
                                saveToken = token.value === '//';
                                break;
                            case xpLexer_1.CharLevelState.sep:
                                // keep '/' char if it's the last part of the reversed path meaning we're on an absolute path
                                exitLoop = token.value !== '/';
                                if (!exitLoop && i === 0) {
                                    cleanedTokens.push(token);
                                }
                                if (!exitLoop) {
                                    finalSlashToken = token;
                                    lastTokenWasSlash = true;
                                }
                                break;
                            case xpLexer_1.CharLevelState.lPr:
                            case xpLexer_1.CharLevelState.lB:
                                xpathStacksChecked++;
                                break;
                            default:
                                exitLoop = true;
                                break;
                        }
                        break;
                    case xpLexer_1.TokenLevelState.nodeNameTest:
                    case xpLexer_1.TokenLevelState.attributeNameTest:
                        saveToken = true;
                        break;
                    case xpLexer_1.TokenLevelState.axisName:
                        saveToken = token.value !== 'child';
                        if (!hasParentAxis) {
                            hasParentAxis = ['parent', 'ancestor', 'ancestor-or-self', 'following-sibling', 'preceding-sibling'].indexOf(token.value) !== -1;
                        }
                        break;
                    case xpLexer_1.TokenLevelState.nodeType:
                        saveToken = ['*', '..', 'node', 'element', 'item'].indexOf(token.value) !== -1;
                        if (!hasParentAxis) {
                            hasParentAxis = token.value === '..';
                        }
                        // TODO: for '..' case, we can just pop cleanedTokens and not save?
                        exitLoop = !(saveToken || token.value === '.');
                        break;
                    case xpLexer_1.TokenLevelState.variable:
                        // e.g. $dev/body
                        foundVariableToken = true;
                        let xpv = xpathVariables.find(v => v.token.value === token.value);
                        if (!xpv) {
                            const tv = token.value.substring(1);
                            xpv = xsltVariables.find(v => v.name === tv);
                        }
                        if (!xpv) {
                            const tv = token.value.substring(1);
                            const gd = globalInstructionData.find(d => d.type === xslLexer_1.GlobalInstructionType.Variable && d.name === tv);
                            if (gd === null || gd === void 0 ? void 0 : gd.idNumber) {
                                xpv = { index: gd.idNumber, token: gd.token, name: gd.name };
                            }
                        }
                        if (xpv) {
                            const lastTokenIndex = XsltSymbolProvider.fetchXPathVariableTokens(tokens, xpv);
                            // recursive call:
                            const result = XsltSymbolProvider.filterPathTokens(tokens, globalInstructionData, xsltVariables, xpathVariables, lastTokenIndex, xpathStack);
                            // console.log('xpv', xpv);
                            // console.log('lastTokenIndex', lastTokenIndex, tokens[lastTokenIndex]);
                            // console.log('filteredVarTokens', result.cleanedTokens);
                            if (!hasParentAxis) {
                                hasParentAxis = result.hasParentAxis;
                            }
                            cleanedTokens = cleanedTokens.concat(result.cleanedTokens);
                        }
                        else {
                            // add the $dev token
                            const varFromExtXPath = xsltTokenCompletions_1.XsltTokenCompletions.extXPathVariables.get(token.value.substring(1));
                            if (varFromExtXPath) {
                                const xpLexer = new xpLexer_1.XPathLexer();
                                const lexPosition = { line: 0, startCharacter: 0, documentOffset: 0 };
                                const extTokens = xpLexer.analyse(varFromExtXPath, xpLexer_1.ExitCondition.None, lexPosition);
                                const result = XsltSymbolProvider.filterPathTokens(extTokens, globalInstructionData, xsltVariables, xpathVariables, extTokens.length - 1, []);
                                if (!hasParentAxis) {
                                    hasParentAxis = result.hasParentAxis;
                                }
                                cleanedTokens = cleanedTokens.concat(result.cleanedTokens);
                            }
                            else {
                                cleanedTokens.push(token);
                                exitLoop = true;
                            }
                        }
                    default:
                        exitLoop = true;
                        break;
                }
            }
            if (foundVariableToken) {
                // $var gives context so don't check if we're in a predicate
                break;
            }
            else if (exitLoop) {
                // check to see if we're still within a predicate: el1[el2 and el3]
                exitLoop = false;
                let predicateStart = -1;
                const xpathStackEnd = xpathStack.length - (xpathStacksChecked + 1);
                for (let x = xpathStackEnd; x > -1; x--) {
                    xpathStacksChecked++;
                    const item = xpathStack[x];
                    if ((item.token.charType === xpLexer_1.CharLevelState.lPr || item.token.charType === xpLexer_1.CharLevelState.lB) && item.tokenIndex) {
                        predicateStart = item.tokenIndex;
                        break;
                    }
                }
                if (predicateStart === -1) {
                    if (lastTokenWasSlash2 && finalSlashToken) {
                        cleanedTokens.push(finalSlashToken);
                    }
                    break;
                }
                else {
                    // warning: changes the loop counter which will be decremented next!!
                    i = predicateStart;
                }
            }
            else if (nesting === 0 && saveToken && bracketTokens.length === 0) {
                cleanedTokens.push(token);
            }
            lastTokenWasSlash2 = lastTokenWasSlash;
            lastTokenWasSlash = false;
            saveToken = false;
        }
        return { cleanedTokens, hasParentAxis };
    }
    static fetchXPathVariableTokens(tokens, xpv) {
        let result = -1;
        let nesting = 0;
        let exitForLoop = false;
        let onXsltTokensAtStart = false;
        const startPosiiton = xpv.index + 2;
        for (let index = startPosiiton; index < tokens.length; index++) {
            const token = tokens[index];
            const isXSLToken = token.tokenType >= xsltTokenCompletions_1.XsltTokenCompletions.xsltStartTokenNumber;
            if (isXSLToken) {
                if (index === startPosiiton) {
                    onXsltTokensAtStart = true;
                }
                if (onXsltTokensAtStart) {
                    result++;
                    continue;
                }
                else {
                    break;
                }
            }
            else {
                onXsltTokensAtStart = false;
            }
            let xpathCharType = token.charType;
            let xpathTokenType = token.tokenType;
            switch (xpathTokenType) {
                case xpLexer_1.TokenLevelState.operator:
                    if (xpathCharType === xpLexer_1.CharLevelState.lPr) {
                        nesting++;
                    }
                    else if (xpathCharType === xpLexer_1.CharLevelState.rPr) {
                        nesting--;
                    }
                    else if (nesting === 0 && xpathCharType === xpLexer_1.CharLevelState.sep && token.value === ',') {
                        exitForLoop = true;
                    }
                    break;
                case xpLexer_1.TokenLevelState.complexExpression:
                    if (nesting === 0) {
                        exitForLoop = true;
                    }
                    break;
                default:
                    break;
            }
            if (exitForLoop) {
                break;
            }
            else {
                result++;
            }
        }
        return result + startPosiiton;
    }
    static getExpectedForXPathLocation(tokens, symbols, hasParentAxis) {
        if (symbols.length === 0) {
            return [[], []];
        }
        if (tokens.length === 0) {
            const token = {
                line: 1,
                startCharacter: 0,
                length: 1,
                value: '//',
                tokenType: xpLexer_1.TokenLevelState.operator,
                charType: xpLexer_1.CharLevelState.dSep
            };
            tokens.push(token);
        }
        else {
            const tv = tokens[tokens.length - 1].value;
            if (!(tv === '/' || tv === '//')) {
                const token = {
                    line: 1,
                    startCharacter: 0,
                    length: 1,
                    value: '//',
                    tokenType: xpLexer_1.TokenLevelState.operator,
                    charType: xpLexer_1.CharLevelState.dSep
                };
                tokens.push(token);
            }
        }
        const lastTokenIndex = tokens.length - 1;
        // start with root element symbol:
        let currentSymbols = [];
        let isDocumentNode = false;
        let nextAxis = AxisType.Child;
        let isPathStart = true;
        let isPathEnd = false;
        let unionElementNames = [];
        // track backwards to get all tokens in path
        for (let i = lastTokenIndex; i > -1; i--) {
            const token = tokens[i];
            const tv = token.value;
            let xpathCharType = token.charType;
            let xpathTokenType = token.tokenType;
            let currentAxis = nextAxis;
            isPathEnd = i === 0;
            nextAxis = AxisType.Child;
            let nextSymbols = [];
            let descendantAttrNames = [];
            switch (xpathTokenType) {
                case xpLexer_1.TokenLevelState.operator:
                    switch (xpathCharType) {
                        case xpLexer_1.CharLevelState.sep:
                            if (isPathStart) {
                                isDocumentNode = token.value === '/';
                                nextSymbols = [symbols[0]];
                            }
                            break;
                        case xpLexer_1.CharLevelState.dSep:
                            if (token.value === '//') {
                                // need to just hold and wait for next token
                                // 1. //*
                                // 2. //element
                                nextSymbols = isPathStart ? [symbols[0]] : currentSymbols;
                                const descendantOrSelf = isPathStart;
                                const { outSymbols, attrNameArray } = XsltSymbolProvider.getDescendantSymbols(nextSymbols, descendantOrSelf, hasParentAxis);
                                nextSymbols = outSymbols;
                                descendantAttrNames = attrNameArray;
                                nextAxis = AxisType.Descendant;
                            }
                    }
                    break;
                case xpLexer_1.TokenLevelState.nodeNameTest:
                    if (isDocumentNode) {
                        isDocumentNode = false;
                        // starting point: assume root element
                        if (currentSymbols[0].name === token.value) {
                            nextSymbols = currentSymbols;
                        }
                    }
                    else if (token.union) {
                        unionElementNames.push(token.value);
                        nextSymbols = currentSymbols;
                    }
                    else if (currentAxis !== AxisType.Child && currentAxis !== AxisType.Parent) {
                        nextSymbols = currentSymbols.filter(current => current.name === token.value);
                    }
                    else {
                        unionElementNames.push(token.value);
                        for (let i = 0; i < currentSymbols.length; i++) {
                            const symbol = currentSymbols[i];
                            if (hasParentAxis) {
                                symbol.children.forEach((child) => {
                                    if (child.kind !== vscode.SymbolKind.Array && unionElementNames.indexOf(child.name) !== -1) {
                                        if (!child.parent) {
                                            child.parent = symbol;
                                        }
                                        nextSymbols.push(child);
                                    }
                                });
                            }
                            else {
                                const next = symbol.children.filter(child => child.kind !== vscode.SymbolKind.Array && unionElementNames.indexOf(child.name) !== -1);
                                nextSymbols = nextSymbols.concat(next);
                            }
                        }
                        unionElementNames.length = 0;
                    }
                    break;
                case xpLexer_1.TokenLevelState.attributeNameTest:
                    const attrTokenValue = token.value.substring(1);
                    currentSymbols.forEach(symbol => {
                        const attrContainer = symbol.children.length > 0 ? symbol.children[0] : undefined;
                        if (attrContainer && attrContainer.kind === vscode.SymbolKind.Array) {
                            const attributes = attrContainer.children;
                            attributes.forEach((attr) => {
                                if (attr.name === attrTokenValue) {
                                    attr.isAttr = true;
                                    if (hasParentAxis) {
                                        attr.parent = symbol;
                                    }
                                    nextSymbols.push(attr);
                                }
                            });
                        }
                    });
                    // leave axisType as child
                    break;
                case xpLexer_1.TokenLevelState.axisName:
                    switch (token.value) {
                        case 'self':
                            nextAxis = AxisType.Self;
                            nextSymbols = currentSymbols;
                            break;
                        case 'descendant-or-self':
                        case 'descendant':
                            nextSymbols = isPathStart ? [symbols[0]] : currentSymbols;
                            nextAxis = AxisType.Descendant;
                            if (token.value === 'descendant-or-self') {
                                nextAxis = AxisType.DescendantOrSelf;
                            }
                            const { outSymbols, attrNameArray } = XsltSymbolProvider.getDescendantSymbols(nextSymbols, nextAxis === AxisType.DescendantOrSelf, hasParentAxis);
                            nextSymbols = outSymbols;
                            descendantAttrNames = attrNameArray;
                            break;
                        case 'attribute':
                            currentSymbols.forEach(symbol => {
                                const attrContainer = symbol.children.length > 0 ? symbol.children[0] : undefined;
                                if (attrContainer && attrContainer.kind === vscode.SymbolKind.Array) {
                                    const attributes = attrContainer.children;
                                    attributes.forEach((attr) => {
                                        attr.isAttr = true;
                                        if (hasParentAxis) {
                                            attr.parent = symbol;
                                        }
                                        nextSymbols.push(attr);
                                    });
                                }
                            });
                            nextAxis = AxisType.Attribute;
                            break;
                        case 'parent':
                            currentSymbols.forEach(symbol => { if (symbol.parent)
                                nextSymbols.push(symbol.parent); });
                            nextAxis = AxisType.ParentKeep;
                            break;
                        case 'following-sibling':
                            nextAxis = AxisType.FollowingSibling;
                        case 'preceding-sibling':
                            // include all siblings
                            const isFollowing = nextAxis === AxisType.FollowingSibling;
                            currentSymbols.forEach(symbol => {
                                const sParent = symbol.parent;
                                if (sParent) {
                                    const pos = symbol.range.start;
                                    sParent.children.forEach((c) => {
                                        if (c.kind !== vscode.SymbolKind.Array && ((isFollowing && c.range.start.isAfter(pos) ||
                                            (!isFollowing && c.range.start.isBefore(pos))))) {
                                            if (hasParentAxis && !c.parent) {
                                                c.parent = sParent;
                                            }
                                            nextSymbols.push(c);
                                        }
                                    });
                                }
                            });
                            if (!isFollowing) {
                                nextAxis = AxisType.PrecedingSibling;
                            }
                            break;
                        case 'ancestor-or-self':
                            nextAxis = AxisType.AncestorOrSelf;
                        case 'ancestor':
                            const includeSelf = nextAxis === AxisType.AncestorOrSelf;
                            currentSymbols.forEach(symbol => {
                                let s = symbol;
                                if (includeSelf) {
                                    nextSymbols.push(s);
                                }
                                while (s) {
                                    s = s.parent;
                                    if (s) {
                                        nextSymbols.push(s);
                                    }
                                }
                            });
                            if (!includeSelf) {
                                nextAxis = AxisType.Ancestor;
                            }
                            break;
                    }
                    break;
                case xpLexer_1.TokenLevelState.nodeType:
                    if (isPathStart) {
                        // TODO: start with all descendants?
                        currentSymbols = [symbols[0]];
                    }
                    if (['*', 'node', 'element', 'item'].indexOf(token.value) !== -1) {
                        if (isDocumentNode) {
                            isDocumentNode = false;
                            nextSymbols = currentSymbols;
                        }
                        else if (currentAxis !== AxisType.Child && currentAxis !== AxisType.Parent) {
                            nextSymbols = currentSymbols;
                        }
                        else {
                            currentSymbols.forEach(symbol => {
                                if (hasParentAxis) {
                                    symbol.children.forEach((child) => {
                                        if (child.kind !== vscode.SymbolKind.Array) {
                                            if (!child.parent) {
                                                child.parent = symbol;
                                            }
                                            nextSymbols.push(child);
                                        }
                                    });
                                }
                                else {
                                    const elementChildren = symbol.children.filter(child => child.kind !== vscode.SymbolKind.Array);
                                    nextSymbols = nextSymbols.concat(elementChildren);
                                }
                            });
                        }
                    }
                    else if (token.value === '..') {
                        currentSymbols.forEach(symbol => { if (symbol.parent)
                            nextSymbols.push(symbol.parent); });
                        nextAxis = AxisType.Parent;
                    }
                    break;
                case xpLexer_1.TokenLevelState.variable:
                    // can only happen at start of path - assume all descendant-or-self
                    const descendantOrSelf = true;
                    nextSymbols = [symbols[0]];
                    const { outSymbols, attrNameArray } = XsltSymbolProvider.getDescendantSymbols(nextSymbols, descendantOrSelf, hasParentAxis);
                    nextSymbols = outSymbols;
                    descendantAttrNames = attrNameArray;
                    nextAxis = AxisType.DescendantOrSelf;
                    break;
                case xpLexer_1.TokenLevelState.operator:
                    break;
            }
            currentSymbols = nextSymbols;
            isPathStart = false;
            if (i === 0) {
                let elementNames = new Set();
                let attrNames = new Set();
                if (isDocumentNode) {
                    elementNames.add(currentSymbols[0].name);
                }
                else if (nextAxis === AxisType.Attribute) {
                    currentSymbols.forEach(current => {
                        if (current.isAttr) {
                            attrNames.add(current.name);
                        }
                    });
                }
                else if (nextAxis !== AxisType.Child && nextAxis !== AxisType.Parent) {
                    currentSymbols.forEach(current => {
                        if (current.isAttr) {
                            attrNames.add(current.name);
                        }
                        else {
                            elementNames.add(current.name);
                        }
                    });
                }
                else {
                    for (let x = 0; x < currentSymbols.length; x++) {
                        const curentSymbol = currentSymbols[x];
                        curentSymbol.children.forEach(child => {
                            if (child.kind === vscode.SymbolKind.Array && child.name === 'attributes') {
                                child.children.forEach(child => attrNames.add('@' + child.name));
                            }
                            else {
                                elementNames.add(child.name);
                            }
                        });
                    }
                }
                const attrNameArray = (nextAxis === AxisType.Descendant || nextAxis === AxisType.DescendantOrSelf)
                    ? descendantAttrNames : [...attrNames];
                return [[...elementNames], attrNameArray];
            } // end if (i == 0)
        }
        return [[], []];
    }
    static getDescendantSymbols(currentSymbols, descendantOrSelf, hasParentAxis, nodeName) {
        let newSymbols = [];
        let parentSymbols = [];
        const isNameTest = nodeName !== undefined;
        let outSymbols = descendantOrSelf ? [...currentSymbols] : [];
        let attrNames = new Set();
        do {
            // recursive descent to get all descendants
            currentSymbols.forEach(symbol => {
                symbol.children.forEach((child) => {
                    if (child.kind === vscode.SymbolKind.Array) {
                        child.children.forEach((attr) => {
                            attrNames.add('@' + attr.name);
                            attr.isAttr = true;
                            if (hasParentAxis) {
                                attr.parent = symbol;
                            }
                        });
                    }
                    else {
                        if (hasParentAxis) {
                            child.parent = symbol;
                        }
                        outSymbols.push(child);
                    }
                });
            });
            newSymbols = newSymbols.concat(outSymbols);
            currentSymbols = outSymbols;
            outSymbols = [];
        } while (currentSymbols.length > 0);
        if (isNameTest) {
            outSymbols = newSymbols.filter(symbol => symbol.name === nodeName);
        }
        else {
            outSymbols = newSymbols;
        }
        const attrNameArray = [...attrNames];
        return { outSymbols, attrNameArray };
    }
    static getSymbolFromXPathLocator(rawText, symbols) {
        const text = rawText.startsWith('/') ? rawText.substring(1) : '';
        const pathParts = text.split('/');
        if (symbols.length === 0) {
            return;
        }
        let currentSymbol = symbols[0];
        pathParts.forEach((item, i) => {
            const isLastItem = i === pathParts.length - 1;
            if (currentSymbol === undefined) {
                // can't yet handle attribute-test: /element/@attribute
                return;
            }
            else if (isLastItem && item.startsWith('@') && currentSymbol.children.length > 0) {
                const attrName = item.substring(1);
                const attributesSymbol = undefined;
                const firstChild = currentSymbol.children[0];
                if (firstChild.kind === vscode.SymbolKind.Array && firstChild.name === 'attributes') {
                    currentSymbol = firstChild.children.find(symbol => symbol.name === attrName);
                }
                else {
                    currentSymbol = undefined;
                }
                return;
            }
            const pos = item.indexOf('[');
            let pathName = '';
            // only want one result;
            let pathIndex = 1;
            if (pos === -1) {
                pathName = item;
            }
            else {
                const predicateString = item.substring(pos + 1, item.length - 1);
                const predicateNum = Number(predicateString);
                pathIndex = -1;
                if (predicateNum !== NaN) {
                    pathIndex = predicateNum;
                }
                pathName = item.substr(0, pos);
            }
            if (i === 0) {
                const spacePos = currentSymbol.name.indexOf(' ');
                const symbolName = spacePos === -1 ? currentSymbol.name : currentSymbol.name.substring(0, spacePos);
                currentSymbol = symbolName === pathName && pathIndex === 1 ? currentSymbol : undefined;
            }
            else {
                let nameCount = 0;
                currentSymbol = currentSymbol.children.find((symbol) => {
                    const spacePos = symbol.name.indexOf(' ');
                    const symbolName = spacePos === -1 ? symbol.name : symbol.name.substring(0, spacePos);
                    if (symbolName === pathName) {
                        nameCount++;
                        return pathIndex === nameCount;
                    }
                    else {
                        return false;
                    }
                });
            }
        });
        return currentSymbol;
    }
    static processImportedGlobals(xsltPackages, importedGlobals1, level1Hrefs, topLevel) {
        return __awaiter(this, void 0, void 0, function* () {
            let level2Globals = [];
            let level2Hrefs = XsltSymbolProvider.accumulateImportHrefs(xsltPackages, importedGlobals1, level1Hrefs);
            let newGlobals = [];
            level2Hrefs.forEach((href) => {
                level2Globals.push(XsltSymbolProvider.fetchImportedGlobals([href]));
            });
            let importedGlobals2Array = yield Promise.all(level2Globals);
            importedGlobals2Array.forEach((importedGlobals2) => {
                if (topLevel) {
                    newGlobals = newGlobals.concat(importedGlobals2);
                }
                else {
                    importedGlobals1 = importedGlobals1.concat(importedGlobals2);
                }
            });
            if (topLevel) {
                return { globals: newGlobals, hrefs: level2Hrefs };
            }
            else {
                return { globals: importedGlobals1, hrefs: level2Hrefs };
            }
        });
    }
    static accumulateImportHrefs(xsltPackages, importedGlobals, existingHrefs) {
        let result = [];
        const rootPath = vscode.workspace.rootPath;
        importedGlobals.forEach((importedG) => {
            importedG.data.forEach((data) => {
                if (data.type === xslLexer_1.GlobalInstructionType.Import || data.type === xslLexer_1.GlobalInstructionType.Include) {
                    let resolvedName = XsltSymbolProvider.resolvePath(data.name, importedG.href);
                    if (existingHrefs.indexOf(resolvedName) < 0) {
                        existingHrefs.push(resolvedName);
                        result.push(resolvedName);
                    }
                }
                else if (rootPath && data.type === xslLexer_1.GlobalInstructionType.UsePackage) {
                    let packageLookup = xsltPackages.find((pkg) => {
                        return pkg.name === data.name;
                    });
                    if (packageLookup) {
                        let resolvedName = XsltSymbolProvider.resolvePathInSettings(packageLookup.path, rootPath);
                        if (existingHrefs.indexOf(resolvedName) < 0) {
                            existingHrefs.push(resolvedName);
                            result.push(resolvedName);
                        }
                    }
                }
            });
        });
        return result;
    }
    static getImportHrefs(xsltPackages, importedGlobals) {
        let result = [];
        const rootPath = vscode.workspace.rootPath;
        importedGlobals.forEach((importedG) => {
            importedG.data.forEach((data) => {
                if (data.type === xslLexer_1.GlobalInstructionType.Import || data.type === xslLexer_1.GlobalInstructionType.Include) {
                    let resolvedName = XsltSymbolProvider.resolvePath(data.name, importedG.href);
                    result.push(resolvedName);
                }
                else if (rootPath && data.type === xslLexer_1.GlobalInstructionType.UsePackage) {
                    let packageLookup = xsltPackages.find((pkg) => {
                        return pkg.name === data.name;
                    });
                    if (packageLookup) {
                        let resolvedName = XsltSymbolProvider.resolvePathInSettings(packageLookup.path, rootPath);
                        result.push(resolvedName);
                    }
                }
            });
        });
        return result;
    }
    static resolvePath(href, documentPath) {
        if (path.isAbsolute(href)) {
            return href;
        }
        else if (href.startsWith('file:///')) {
            return href.substring(7);
        }
        else {
            href = href.startsWith('file:') ? href.substring(5) : href;
            let basePath = path.dirname(documentPath);
            let joinedPath = path.join(basePath, href);
            return path.normalize(joinedPath);
        }
    }
    static resolvePathInSettings(href, workspace) {
        if (path.isAbsolute(href)) {
            return href;
        }
        else {
            let joinedPath = path.join(workspace, href);
            return path.normalize(joinedPath);
        }
    }
    static fetchImportedGlobals(inputHrefs) {
        return __awaiter(this, void 0, void 0, function* () {
            let result = [];
            //let inputHrefs: string[] = this.accumulateImportHrefs(globalInstructionData, existingHrefs, docHref);
            let lastIndex = inputHrefs.length - 1;
            if (lastIndex < 0) {
                return result;
            }
            else {
                return new Promise((resolve, reject) => {
                    inputHrefs.forEach((href, index) => {
                        globalsProvider_1.GlobalsProvider.provideGlobals(href).then((globals) => {
                            result.push({ href: href, data: globals.data, error: globals.error });
                            if (index === lastIndex) {
                                resolve(result);
                            }
                        });
                    });
                });
            }
        });
    }
}
exports.XsltSymbolProvider = XsltSymbolProvider;
XsltSymbolProvider.documentSymbols = new Map();
XsltSymbolProvider.importSymbolHrefs = new Map();
XsltSymbolProvider.useSourceFile = vscode.workspace.getConfiguration('XSLT.resources').get('useSourceFile');
//# sourceMappingURL=xsltSymbolProvider.js.map